# -*- encoding: utf-8 -*-

"""
Code for dealing with quotes.

This file is part of the wpm software.
Copyright 2017, 2018 Christian Stigen Larsen

Distributed under the GNU Affero General Public License (AGPL) v3 or later. See
the file LICENSE.txt for the full license text. This software makes use of open
source software.

The quotes database is *not* covered by the AGPL!
"""

import codecs
import gzip
import json
import os
import random
import sys

import pkg_resources

from wpm.error import WpmError

class Quote(object):
    """Holds a single quote."""
    # pylint: disable=too-few-public-methods

    def __init__(self, author, title, text, text_id):
        self.author = author
        self.text = text
        self.title = title
        self.text_id = text_id

    @staticmethod
    def from_tuple(data):
        """Converts tuple of (author, text, title, text_id) to Quote object."""
        author = data[0]
        title = data[1]
        text = data[2]
        text_id = data[3]
        return Quote(author, title, text, text_id)

    def __str__(self):
        return "\"%s\"\n\t- %s: %s" % (self.text, self.author, self.title)


class RandomIterator(object):
    """Random, bi-directional iterator."""
    def __init__(self, quotes):
        self.quotes = quotes
        self.indices = list(range(len(self.quotes)))
        self.index = 0
        random.shuffle(self.indices)

    def __len__(self):
        return len(self.quotes)

    def current(self):
        """Returns current quote."""
        index = self.indices[self.index]
        return self._get_quote(index)

    def __getitem__(self, index):
        return self._get_quote(index)

    def _get_quote(self, index):
        quote = self.quotes[index]
        author = quote[0]
        title = quote[1]
        text = quote[2]
        if len(quote) > 3:
            text_id = quote[3]
        else:
            text_id = index
        return Quote(author, title, text, text_id)

    def put_to_front(self, text_ids):
        """Puts given text IDs to the very front of the queue."""
        front = []
        back = []

        for index in range(len(self.quotes)):
            quote = self[index]

            if quote.text_id in text_ids:
                front.append(index)
            else:
                back.append(index)

        random.shuffle(back)
        self.indices = front + back
        self.index = 0

    @property
    def database(self):
        """The quotes database."""
        return self.quotes.database

    @property
    def text_id(self):
        """Returns text ID of current quote."""
        return self.indices[self.index]

    def next(self):
        """Goes to next quote."""
        self.index = (self.index + 1) % len(self.quotes)
        return self.current()

    def previous(self):
        """Goes back to previous quote."""
        self.index = (self.index - 1) % len(self.quotes)
        return self.current()


class Quotes(object):
    """Container for quotes."""
    def __init__(self, quotes=None, database=None):
        self.quotes = quotes
        self.database = database
        self._index = None

    def _build_index(self):
        """Builds an index of text_ids."""
        if self._index is None:
            self._index = {}
            for index, quote in enumerate(self.quotes):
                quote = Quote.from_tuple(quote)
                self._index[quote.text_id] = index

    def from_id(self, text_id):
        """Looks up quote from text_id."""
        self._build_index()
        index = self._index[text_id]
        return Quote.from_tuple(self.quotes[index])

    def __len__(self):
        return len(self.quotes)

    def __getitem__(self, index):
        return self.quotes[index]

    def at(self, index):
        return Quote.from_tuple(self[index])

    def __setitem__(self, index, item):
        if not isinstance(item, list):
            raise ValueError("Expected a list")
        if len(item) != 3:
            raise ValueError("Expected a list of three strings")
        self.quotes[index] = item

    def random_iterator(self):
        """Returns a random iterator for all the quotes."""
        return RandomIterator(self)

    @staticmethod
    def _database_filename():
        """Returns the filename of the packaged database."""
        return pkg_resources.resource_filename("wpm", "data/examples.json.gz")

    @staticmethod
    def load_json(filename=None):
        """Loads quotes from a JSON file."""
        if filename is None:
            filename = Quotes._database_filename()

        try:
            with codecs.open(filename, encoding="utf-8") as file_obj:
                quotes = json.load(file_obj)
        except Exception as error:
            raise WpmError("Could not read JSON file: %s" % error)

        # Flatten
        out = []
        for text_id, quote in enumerate(quotes):
            author = quote["author"]
            title = quote["title"]
            text = quote["text"]
            text_id = int(quote.get("id", text_id))
            out.append((author, title, text, text_id))

        return Quotes(list(set(out)))

    @staticmethod
    def load(filename=None):
        """Loads quotes from gzipped JSON file."""
        if filename is None:
            filename = Quotes._database_filename()
            database = "default"
        else:
            database = os.path.splitext(os.path.basename(filename))[0]

        args = {"filename": filename, "mode": "rt"}
        if sys.version_info.major == 3:
            args["encoding"] = "utf-8"

        with gzip.open(**args) as file_obj:
            quotes = json.load(file_obj)
            quotes = tuple(map(tuple, quotes))
            return Quotes(quotes, database)

    def save(self, filename=None):
        """Saves current quotes to gzipped JSON file."""
        if filename is None:
            filename = Quotes._database_filename()

        args = {"filename": filename, "mode": "wt"}
        if sys.version_info.major == 3:
            args["encoding"] = "utf-8"

        with gzip.open(**args) as file_obj:
            json.dump(self.quotes, file_obj)



"""
Functions for dealing with the normal distribution.
"""

import math

def memoized(func):
    """A memoization decorator."""
    cache = {}
    def wrap(*args):
        if args in cache:
            return cache[args]
        result = func(*args)
        cache[args] = result
        return result
    return wrap

@memoized
def c(k):
    """The MacLaurin coefficient for erf_inv."""
    if k == 0:
        return 1

    res = 0
    for m in range(k):
        res += (c(m) * c(k - 1 - m)) / float((m + 1) * (2*m + 1))

    return res

def erf_inv(z, steps=10):
    """THe inverse error function approximated through its MacLaurin series
    expansion.

    The value z must be inside <-1, 1>.
    """
    if not (-1 < z < 1):
        raise ValueError(z)

    res = 0
    for k in range(steps):
        res += (c(k) / (2*k + 1)) * (math.sqrt(math.pi) * z / 2.0)**(2*k + 1)

    return res

def phi_inv(p):
    """The normal distribution's quantile function z_p / phi^(-1)(p)."""
    return math.sqrt(2) * erf_inv(2*p - 1)

def confidence_interval(mu, sd, n, alpha):
    """Calculates the confidence interval given the normal distribution."""
    if n == 0:
        return 0, 0
    z = phi_inv(1.0 - alpha/2.0)
    return mu - z*sd/math.sqrt(n), mu + z*sd/math.sqrt(n)



# -*- encoding: utf-8 -*-

"""
This file is part of the wpm software.
Copyright 2017, 2018 Christian Stigen Larsen

Distributed under the GNU Affero General Public License (AGPL) v3 or later. See
the file LICENSE.txt for the full license text. This software makes use of open
source software.

The quotes database is *not* covered by the AGPL!
"""

import argparse
import codecs
import os
import sys

from wpm.gauss import confidence_interval
import wpm
import wpm.config
import wpm.error
import wpm.game
import wpm.quotes
import wpm.stats

def parse_args():
    """Parses command line arguments."""
    argp = argparse.ArgumentParser(prog="wpm", epilog=wpm.__copyright__)

    argp.add_argument("--load-json", metavar="FILENAME", default=None,
                      help="""JSON file containing quotes.

The format is

[{"author": "...", "title": "...", "text": "...", "id": ...}, ...]
""")
    argp.add_argument("--load", metavar="FILENAME", default=None,
                      help="A pure text file to train on.")

    argp.add_argument("-V", "--version", default=False, action="store_true",
                      help="Show program version")

    argp.add_argument("--keyboard", default=None, type=str,
                      help="Records WPM statistics under the given keyboard name")

    argp.add_argument("-s", "--stats", default=False, action="store_true",
                      help="Shows keyboard statistics")

    argp.add_argument("--stats-file", default="~/.wpm.csv", type=str,
                      help="File to record score history to (CSV format)")

    argp.add_argument("--id", "-i", default=None, type=int,
                      help="If specified, jumps to given text ID on start.")

    argp.add_argument("--search", default=None, type=str,
                      help="Put quotes/authors/titles matching case-insensitive text query first")

    opts = argp.parse_args()

    if opts.version:
        print("WPM v%s" % wpm.__version__)
        print(wpm.__copyright__)
        print("Source code (sans quotes) distributed under the %s" % wpm.__license__)
        sys.exit(0)

    opts.stats_file = os.path.expanduser(opts.stats_file)
    return opts

def load_stats(filename, keyboard):
    """Loads CSV stats from file."""
    if not os.path.isfile(filename):
        return wpm.stats.Stats(keyboard)

    try:
        stats = wpm.stats.Stats.load(filename)
    except ValueError:
        new_name = filename + ".old"
        print("Unsupported format. Renaming %s to %s" % (filename,
                                                         new_name))
        os.rename(filename, new_name)
        stats = wpm.stats.Stats()

    if keyboard is not None:
        stats.keyboard = keyboard

    return stats

def load_json_quotes(filename):
    """Loads quotes from JSON file."""
    if filename is not None:
        return wpm.quotes.Quotes.load_json(filename)
    return []

def load_plain_text_quote(quotes, filename):
    """Loads quotes from plain text file."""
    if not os.path.isfile(filename):
        raise wpm.error.WpmError("No such file: %s" % filename)

    with codecs.open(filename, encoding="utf-8") as file_obj:
        text = file_obj.read()
        text = text.replace("\r", "").rstrip()

        author = ""
        title = os.path.basename(filename)
        database = "plaintext"

        # For now, just use inode for the text ID
        text_id = os.stat(filename).st_ino

        quotes.append([author, title, text, text_id])
        return wpm.quotes.Quotes(quotes, database=database)

def print_stats(stats):
    """Prints table of game results."""
    table = []

    config = wpm.config.Config()
    percent = config.confidence_interval_percent

    for keyboard in sorted(stats.games.keys()):
        name = keyboard if keyboard is not None else "n/a"

        for last_n in [0, 10, 50, 100, 500, 1000]:
            results = stats.results(keyboard, last_n=last_n)

            if len(results) >= last_n:
                if last_n == 0:
                    label = len(results)
                else:
                    label = "n-%d" % last_n

                wpm_avg, acc_avg = results.averages()
                wpm_sd, acc_sd = results.stddevs()

                alpha = 1 - (percent/100.0)
                wpm_ci = confidence_interval(wpm_avg, wpm_sd, len(results), alpha)

                table.append([name,
                              label,
                              wpm_avg,
                              wpm_sd,
                              100.0*acc_avg,
                              100.0*acc_sd,
                              wpm_ci[0],
                              wpm_ci[1]])

    if table:
        width = max(max(len(e[0]) for e in table), 11)
    else:
        width = 0

    print("Keyboard     Games   WPM                        Accuracy")
    print("                     average stddev %d%% ci      average stddev" % percent)

    for entry in table:
        label, count, wpm_avg, wpm_sd, acc_avg, acc_sd, ci0, ci1 = entry
        print("%-*s   %5s  %5.1f  %5.1f   %5.1f-%5.1f %5.1f%%  %5.1f%%" %
              (width, label, count, wpm_avg, wpm_sd, ci0, ci1, acc_avg, acc_sd))

def search(quotes, query):
    """Returns text IDs for quotes matching query."""
    for quote in iter(quotes):
        quote = wpm.quotes.Quote.from_tuple(quote)

        author = quote.author.lower()
        title = quote.title.lower()
        text = quote.text.lower()

        if (query in text) or (query in author) or (query in title):
            yield quote.text_id

def main():
    """Main entry point for command line invocation."""
    try:
        opts = parse_args()
        stats = load_stats(opts.stats_file, opts.keyboard)

        if opts.load_json is not None:
            quotes = load_json_quotes(opts.load_json)
        elif opts.load is not None:
            quotes = load_plain_text_quote([], opts.load)
        else:
            # Load default database
            quotes = wpm.quotes.Quotes.load()

        if opts.stats:
            print_stats(stats)
            return

        if opts.search:
            text_ids = list(search(quotes, opts.search.lower()))

            if not text_ids:
                print("No quotes matching %r" % opts.search)
                sys.exit(1)
        elif opts.id is not None:
            text_ids = [opts.id]
        else:
            text_ids = []
    except wpm.error.WpmError as error:
        print(error)
        sys.exit(1)

    with wpm.game.GameManager(quotes, stats) as gm:
        try:
            gm.run(to_front=text_ids)
            gm.stats.save(opts.stats_file)
        except KeyboardInterrupt:
            gm.stats.save(opts.stats_file)
            sys.exit(0)



# -*- encoding: utf-8 -*-

"""
This file is part of the wpm software.
Copyright 2017, 2018 Christian Stigen Larsen

Distributed under the GNU Affero General Public License (AGPL) v3 or later. See
the file LICENSE.txt for the full license text. This software makes use of open
source software.

The quotes database is *not* covered by the AGPL!
"""

import wpm.commandline

if __name__ == "__main__":
    wpm.commandline.main()



"""
Code for dealing with user configuration settings (rc-file).

This file is part of the wpm software.
Copyright 2017, 2018 Christian Stigen Larsen

Distributed under the GNU Affero General Public License (AGPL) v3 or later. See
the file LICENSE.txt for the full license text. This software makes use of open
source software.

The quotes database is *not* covered by the AGPL!
"""

import curses
import os

try:
    import configparser
except ImportError:
    # Python 2.7
    import ConfigParser as configparser

class Config(object):
    """Contains the user configuration, backed by the .wpmrc file."""
    # pylint: disable=too-many-public-methods

    config = None

    def __init__(self):
        Config.config = configparser.ConfigParser()
        self.filename = os.path.expanduser("~/.wpmrc")

        if os.path.isfile(self.filename):
            self.load()
        else:
            Config.set_defaults()
            self.save()

    def load(self):
        """Loads ~/.wpmrc config settings."""
        Config.config.read(self.filename)

    def save(self):
        """Saves settings to ~/.wpmrc"""
        with open(self.filename, "wt") as file_obj:
            Config.config.write(file_obj)

    @staticmethod
    def set_defaults():
        """Sets up default settings."""
        Config.config.add_section("curses")
        Config.config.set("curses", "escdelay", "15")
        Config.config.set("curses", "window_timeout", "20")

        Config.config.add_section("wpm")
        Config.config.set("wpm", "confidence_interval_percent", "95")
        Config.config.set("wpm", "max_quote_width", "-1")
        Config.config.set("wpm", "spaces_to_expand_tabs_to", "1")

        Config.config.add_section("xterm-256color")
        Config.config.set("xterm-256color", "author_bg", str(233))
        Config.config.set("xterm-256color", "author_fg", str(240))
        Config.config.set("xterm-256color", "background", str(233))
        Config.config.set("xterm-256color", "correct_bg", str(233))
        Config.config.set("xterm-256color", "correct_fg", str(240))
        Config.config.set("xterm-256color", "incorrect_bg", str(52))
        Config.config.set("xterm-256color", "incorrect_fg", str(197))
        Config.config.set("xterm-256color", "prompt_bg", str(233))
        Config.config.set("xterm-256color", "prompt_fg", str(244))
        Config.config.set("xterm-256color", "quote_bg", str(233))
        Config.config.set("xterm-256color", "quote_fg", str(195))
        Config.config.set("xterm-256color", "score_highlight_bg", str(197))
        Config.config.set("xterm-256color", "score_highlight_fg", str(230))
        Config.config.set("xterm-256color", "status_bg", str(24))
        Config.config.set("xterm-256color", "status_fg", str(51))

        Config.config.add_section("xterm-colors")
        Config.config.set("xterm-colors", "author_bg", str(curses.COLOR_BLACK))
        Config.config.set("xterm-colors", "author_fg", str(curses.COLOR_WHITE))
        Config.config.set("xterm-colors", "background", str(curses.COLOR_BLACK))
        Config.config.set("xterm-colors", "correct_bg", str(curses.COLOR_BLACK))
        Config.config.set("xterm-colors", "correct_fg", str(curses.COLOR_WHITE))
        Config.config.set("xterm-colors", "incorrect_bg", str(curses.COLOR_BLACK))
        Config.config.set("xterm-colors", "incorrect_fg", str(curses.COLOR_RED))
        Config.config.set("xterm-colors", "prompt_bg", str(curses.COLOR_BLACK))
        Config.config.set("xterm-colors", "prompt_fg", str(curses.COLOR_WHITE))
        Config.config.set("xterm-colors", "quote_bg", str(curses.COLOR_BLACK))
        Config.config.set("xterm-colors", "quote_fg", str(curses.COLOR_WHITE))
        Config.config.set("xterm-colors", "score_highlight_bg", str(curses.COLOR_RED))
        Config.config.set("xterm-colors", "score_highlight_fg", str(curses.COLOR_YELLOW))
        Config.config.set("xterm-colors", "status_bg", str(curses.COLOR_BLUE))
        Config.config.set("xterm-colors", "status_fg", str(curses.COLOR_CYAN))

    @property
    def escdelay(self):
        """Curses ESCDELAY."""
        return Config.config.get("curses", "escdelay")

    @property
    def window_timeout(self):
        """Curses window timeout."""
        return int(Config.config.get("curses", "window_timeout"))

    @property
    def max_quote_width(self):
        """Length to wrap quotes at. Inactive if zero or less."""
        try:
            return int(Config.config.get("wpm", "max_quote_width"))
        except configparser.NoSectionError:
            Config.config.add_section("wpm")
            Config.config.set("wpm", "max_quote_width", "-1")
            return -1
        except configparser.NoOptionError:
            Config.config.set("wpm", "max_quote_width", "-1")
            return -1

    @property
    def confidence_interval_percent(self):
        """Used to construct an x% confidence interval."""
        try:
            return int(Config.config.get("wpm", "confidence_interval_percent"))
        except configparser.NoSectionError:
            Config.config.add_section("wpm")
            Config.config.set("wpm", "confidence_interval_percent", "95")
            return 95
        except configparser.NoOptionError:
            Config.config.set("wpm", "confidence_interval_percent", "95")
            return 95

    @property
    def tab_spaces(self):
        """Used to construct an x% confidence interval."""
        try:
            return int(Config.config.get("wpm", "spaces_to_expand_tabs_to"))
        except configparser.NoSectionError:
            Config.config.add_section("wpm")
            Config.config.set("wpm", "spaces_to_expand_tabs_to", "1")
            return 1
        except configparser.NoOptionError:
            Config.config.set("wpm", "spaces_to_expand_tabs_to", "1")
            return 1

    @property
    def background_color_256(self):
        """Background color for 256-color terminalss."""
        return int(Config.config.get("xterm-256color", "background"))

    @property
    def incorrect_color_256(self):
        """Incorrectly type quote color for 256-color terminals."""
        return (int(Config.config.get("xterm-256color", "incorrect_fg")),
                int(Config.config.get("xterm-256color", "incorrect_bg")))

    @property
    def status_color_256(self):
        """Top status bar color for 256-color terminals."""
        return (int(Config.config.get("xterm-256color", "status_fg")),
                int(Config.config.get("xterm-256color", "status_bg")))

    @property
    def correct_color_256(self):
        """Color for correctly typed text for 256-color terminals."""
        return (int(Config.config.get("xterm-256color", "correct_fg")),
                int(Config.config.get("xterm-256color", "correct_bg")))

    @property
    def quote_color_256(self):
        """Quote color for 256-color terminals."""
        return (int(Config.config.get("xterm-256color", "quote_fg")),
                int(Config.config.get("xterm-256color", "quote_bg")))

    @property
    def author_color_256(self):
        """Author color for 256-color terminals."""
        return (int(Config.config.get("xterm-256color", "author_fg")),
                int(Config.config.get("xterm-256color", "author_bg")))

    @property
    def prompt_color_256(self):
        """Prompt, or input, color for 256-color terminals."""
        return (int(Config.config.get("xterm-256color", "prompt_fg")),
                int(Config.config.get("xterm-256color", "prompt_bg")))

    @property
    def score_highlight_color_256(self):
        """Highlighted score color for 256-color displays."""
        return (int(Config.config.get("xterm-256color", "score_highlight_fg")),
                int(Config.config.get("xterm-256color", "score_highlight_bg")))

    @property
    def background_color(self):
        """Background color."""
        return int(Config.config.get("xterm-colors", "background"))

    @property
    def incorrect_color(self):
        """Color of incorrectly typed text."""
        return (int(Config.config.get("xterm-colors", "incorrect_fg")),
                int(Config.config.get("xterm-colors", "incorrect_bg")))

    @property
    def status_color(self):
        """Top status bar color."""
        return (int(Config.config.get("xterm-colors", "status_fg")),
                int(Config.config.get("xterm-colors", "status_bg")))

    @property
    def correct_color(self):
        """Color for correctly typed text."""
        return (int(Config.config.get("xterm-colors", "correct_fg")),
                int(Config.config.get("xterm-colors", "correct_bg")))

    @property
    def quote_color(self):
        """Normal color for quote."""
        return (int(Config.config.get("xterm-colors", "quote_fg")),
                int(Config.config.get("xterm-colors", "quote_bg")))

    @property
    def author_color(self):
        """Color for author."""
        return (int(Config.config.get("xterm-colors", "author_fg")),
                int(Config.config.get("xterm-colors", "author_bg")))

    @property
    def prompt_color(self):
        """Color for the input field."""
        return (int(Config.config.get("xterm-colors", "prompt_fg")),
                int(Config.config.get("xterm-colors", "prompt_bg")))

    @property
    def score_highlight_color(self):
        """Highlighted score color."""
        return (int(Config.config.get("xterm-colors", "score_highlight_fg")),
                int(Config.config.get("xterm-colors", "score_highlight_bg")))



# -*- encoding: utf-8 -*-

"""
This file is part of the wpm software.
Copyright 2017, 2018 Christian Stigen Larsen

Distributed under the GNU Affero General Public License (AGPL) v3 or later. See
the file LICENSE.txt for the full license text. This software makes use of open
source software.

The quotes database is *not* covered by the AGPL!
"""

import collections
import csv
import datetime
import math
import os

class Timestamp(object):
    """Methods for dealing with timestamps."""
    DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S.%f"

    @staticmethod
    def from_string(string):
        """Parses timestamp from string in ``Timestamp.DATETIME_FORMAT``."""
        return datetime.datetime.strptime(string, Timestamp.DATETIME_FORMAT)

    @staticmethod
    def now():
        """Returns current UTC time."""
        return datetime.datetime.utcnow()


class GameResult(object):
    """Contains the result of one game."""
    # pylint: disable=too-many-instance-attributes, too-few-public-methods

    def __init__(self, game):
        race, wpm, accuracy, rank, racers, text_id, timestamp, database = game
        self.race = race
        self.wpm = wpm
        self.accuracy = accuracy
        self.rank = rank
        self.racers = racers
        self.text_id = text_id
        self.timestamp = timestamp
        self.database = database

    def __repr__(self):
        return "<GameResult: %s wpm=%.1f acc=%.1f id=%d>" % (
                self.timestamp, self.wpm, self.accuracy, self.text_id)


class GameResults(object):
    """Container for several GameResult objects."""
    def __init__(self, keyboard, games):
        self.keyboard = keyboard
        self.games = games

    @property
    def results(self):
        """Yields all the ``GameResult`` objects."""
        for game in self.games:
            yield GameResult(game)

    def __repr__(self):
        return "<GameResults: len=%d keyboard=%r>" % (len(self.games),
                                                      self.keyboard)

    def append(self, game):
        self.games.append(game)

    def __len__(self):
        return len(self.games)

    def extremals(self):
        init = 999999
        min_wpm = init
        max_wpm = 0
        min_acc = init
        max_acc = 0

        for result in self.results:
            min_wpm = min(min_wpm, result.wpm)
            max_wpm = max(max_wpm, result.wpm)
            min_acc = min(min_acc, result.accuracy)
            max_acc = max(max_acc, result.accuracy)

        if min_wpm == init:
            min_wpm = 0
        if min_acc == init:
            min_acc = 0

        return min_wpm, max_wpm, min_acc, max_acc

    def averages(self):
        """Returns a tuple of WPM and accuracy averages."""
        if not self.games:
            return 0, 0

        wpms = 0
        accs = 0

        for result in self.results:
            wpms += result.wpm
            accs += result.accuracy

        return wpms / len(self), accs / len(self)

    def stddevs(self):
        """Returns a tuple of WPM and accuracy standard deviations.

        Calculated from the root of the sample variance.
        """
        samples = len(self)

        if samples <= 1:
            return 0.0, 0.0

        wpm_sd = 0
        acc_sd = 0

        wpm_avg, acc_avg = self.averages()

        for result in self.results:
            wpm_sd += (result.wpm - wpm_avg)**2.0
            acc_sd += (result.accuracy - acc_avg)**2.0

        return (math.sqrt(wpm_sd/(samples - 1)),
                math.sqrt(acc_sd/(samples - 1)))


class Stats(object):
    """Typing statistics"""
    def __init__(self, current_keyboard=None, games=None):
        self.keyboard = current_keyboard
        if games is None:
            self.games = collections.defaultdict(list)
        else:
            self.games = games

    def __repr__(self):
        return "<Stats: keyboards=%d current=%r>" % (len(self), self.keyboard)

    def results(self, keyboard=None, last_n=0):
        """Returns the ``GameResults``."""
        if keyboard is None:
            keyboard = self.keyboard
        return GameResults(keyboard, self.games[keyboard][-last_n:])

    def text_id_results(self, keyboard, text_id):
        results = GameResults(self.keyboard, [])

        for game in self.games[keyboard]:
            if game[5] == text_id:
                results.append(game)

        return results

    def add(self, wpm, accuracy, text_id, database):
        """Adds a game result to the stats."""
        race = 0
        rank = 1
        racers = 1

        self.games[self.keyboard].append((
            race,
            wpm,
            accuracy,
            rank,
            racers,
            text_id,
            Timestamp.now(),
            database))

    def average(self, keyboard=None, last_n=None):
        """Returns the average WPM."""
        return self.results(keyboard, last_n).averages()[0]

    def __len__(self):
        return len(self.games)

    def __getitem__(self, key):
        return self.games[key]

    def keys(self):
        """Returns keys for this dict-like object."""
        return self.games.keys()

    def values(self):
        """Returns values for this dict-like object."""
        return self.games.values()

    def items(self):
        """Returns tuple of keys and values for this dict-like object."""
        return self.games.items()

    @staticmethod
    def load(filename=None):
        """Loads stats from a CSV file."""

        if filename is None:
            filename = os.path.expanduser("~/.wpm.csv")

        games = collections.defaultdict(list)
        current_keyboard = None

        def parse(row):
            """Converts CSV row to internal types."""
            race = int(row[0])
            wpm = float(row[1])
            accuracy = float(row[2])
            rank = int(row[3])
            racers = int(row[4])
            text_id = int(row[5])
            timestamp = Timestamp.from_string(row[6])
            database = row[7]
            keyboard = row[8]

            return (race, wpm, accuracy, rank, racers, text_id, timestamp,
                    database, keyboard)

        with open(filename, "rt") as file_obj:
            reader = csv.reader(file_obj)

            for row in reader:
                result = parse(row)
                keyboard = result[-1]
                games[keyboard].append(result[:-1])
                current_keyboard = keyboard

        return Stats(current_keyboard, games)

    def save(self, filename):
        """Writes game results to a CSV file compatible with the one from
        TypeRacer."""
        allgames = []
        for keyboard, games in self.items():
            if keyboard is None:
                keyboard = "Unspecified"

            for game in games:
                allgames.append(list(game) + [keyboard])

        by_time = lambda row: row[6]
        games = sorted(allgames, key=by_time)

        # Write to a temp file just in case we get an exception
        with open(filename + ".tmp", "wt") as file_obj:
            writer = csv.writer(file_obj)

            for race, game in enumerate(games):
                # The timestamp is the race number, so use that instead.
                game[0] = 1 + race
                writer.writerow(game)

        os.rename(filename + ".tmp", filename)



# -*- encoding: utf-8 -*-

"""
Measures your typing speed in words per minute (WPM).

This file is part of the wpm software.
Copyright 2017, 2018 Christian Stigen Larsen

Distributed under the GNU Affero General Public License (AGPL) v3 or later. See
the file LICENSE.txt for the full license text. This software makes use of open
source software.

The quotes database is *not* covered by the AGPL!
"""

import curses
import time

from wpm.config import Config
from wpm.error import WpmError
from wpm.record import Recorder
from wpm.screen import Screen

class GameManager(object):
    """The main game runner."""
    def __init__(self, quotes, stats):
        self.config = Config()
        self.stats = stats
        self.average = self.stats.average(self.stats.keyboard, last_n=10)
        self.tab_spaces = None

        # Stats
        self.position = 0
        self.incorrect = 0
        self.total_incorrect = 0

        self.start = None
        self.stop = None

        self._edit = ""
        self.num_quotes = len(quotes)
        self.quotes = quotes.random_iterator()

        self.screen = Screen()
        self.set_quote(self.quotes.next())

    def __enter__(self):
        return self

    def __exit__(self, error_type, error_value, error_traceback):
        self.screen.deinit()
        if error_type is not None:
            return False
        return self

    def set_tab_spaces(self, spaces):
        """Sets how many spaces a tab should expand to."""
        self.tab_spaces = spaces

    def mark_finished(self):
        """Marks current race as finished."""
        self.stop = time.time()
        self.stats.add(self.wpm(self.elapsed),
                       self.accuracy,
                       self.quote.text_id,
                       self.quotes.database)

        self.average = self.stats.average(self.stats.keyboard, last_n=10)

    def set_quote(self, quote):
        """Sets current quote."""
        self.quote = quote
        self.recorder = Recorder()
        self.screen.set_quote(self.quote)

    @property
    def is_typing(self):
        """Is user currently typing a quote?"""
        return (self.start is not None) and (self.stop is None)

    @property
    def game_done(self):
        """Has user finished a quote?"""
        return (self.start is not None) and (self.stop is not None)

    def run(self, to_front=None):
        """Starts the main game loop."""
        self.set_tab_spaces(self.config.tab_spaces)

        if to_front:
            self.quotes.put_to_front(to_front)
            self.set_quote(self.quotes.current())

        key = None
        while True:
            self.now = time.time()

            head = self.get_stats(self.elapsed)

            if self.is_typing:
                if self.screen.first_key:
                    self.screen.first_key = False
                    self.recorder.reset()
                    self.screen.rerender_race(head)

                self.screen.show_keystroke(head,
                                           self.position,
                                           self.incorrect,
                                           self._edit,
                                           key)
            elif self.game_done:
                self.screen.show_score(head,
                                       self.wpm(self.elapsed),
                                       self.stats)
            else:
                self.screen.show_browser(head, self.stats)

            self.screen.window.refresh()
            key = self.screen.get_key()
            self.handle_key(key)

    def wpm(self, elapsed):
        """Words per minute."""
        if self.start is None:
            return 0
        return min((60.0 * self.position / 5.0) / elapsed, 999)

    def cps(self, elapsed):
        """Characters per second."""
        if self.start is None:
            return 0
        return min(float(self.position) / elapsed, 99)

    @property
    def elapsed(self):
        """Elapsed game round time."""
        if self.start is None:
            # Typing has not started
            return 0
        if self.stop is None:
            # Currently typing
            return self.now - self.start
        # Done typing
        return self.stop - self.start

    @property
    def accuracy(self):
        """Returns typing accuracy."""
        if self.start is None:
            return 0

        length = len(self.quote.text)
        return float(length) / (length + self.total_incorrect)

    def get_stats(self, elapsed):
        """Returns the top-bar stats line."""
        kbd = self.stats.keyboard

        parts = (
            "%5.1f wpm" % self.wpm(elapsed),
            " %4.1f cps" % self.cps(elapsed),
            " %5.2fs" % elapsed,
            " %5.1f%% acc" % (100.0*self.accuracy),
            " %5.1f avg wpm" % self.average,
            " - " + (kbd if kbd is not None else "Unspecified"),
        )

        stat = ""

        for part in parts:
            if (len(stat) + len(part)) <= self.screen.columns:
                stat += part

        return stat

    def reset(self, direction=0):
        """Cancels current game."""
        self.start = None
        self.stop = None

        self.position = 0
        self.incorrect = 0
        self.total_incorrect = 0
        self.cheight = 0

        self._edit = ""
        self.screen.first_key = True
        self.screen.clear_prompt()

        if direction:
            if direction > 0:
                self.set_quote(self.quotes.next())
            else:
                self.set_quote(self.quotes.previous())
            self.screen.clear()

    def resize(self):
        """Handles a resized terminal."""
        max_y, max_x = self.screen.window.getmaxyx()
        self.screen.clear()

        # Check if we have the resizeterm ncurses extension
        if hasattr(curses, "resizeterm"):
            curses.resizeterm(max_y, max_x)
            # An ungetch for KEY_RESIZE will be sent to let others handle it.
            # We'll just pop it off again to prevent endless loops.
            self.screen.get_key()

        self.screen.set_quote(self.quote)

        if self.start is not None and self.stop is None:
            # Resize during typing requires redrawing quote.
            self.screen.update_quote(Screen.COLOR_QUOTE)
            self.screen.update_author()

            if self.position + self.incorrect <= len(self.quote.text):
                for pos in range(self.position + 1):
                    self.screen.highlight_progress(pos, 0)
                for inc in range(self.incorrect + 1):
                    self.screen.highlight_progress(self.position, inc)

    def handle_key(self, key):
        """Dispatches actions based on key and current mode."""
        # TODO: Refactor this mess of a function
        if key is None:
            return

        if key == "KEY_RESIZE":
            self.resize()
            return

        # Browse mode
        if self.start is None or (self.start is not None and
                                  self.stop is not None):
            if key in (" ", "KEY_LEFT", "KEY_RIGHT"):
                self.reset(direction=-1 if key == "KEY_LEFT" else 1)
                return
            elif Screen.is_escape(key):
                # Exit program
                raise KeyboardInterrupt()

        if Screen.is_escape(key):
            self.reset()
            return

        self.recorder.add(self.elapsed, key, self.position, self.incorrect)

        if Screen.is_backspace(key):
            if self.incorrect:
                self.incorrect -= 1
                self._edit = self._edit[:-1]
            elif self._edit:
                self.position -= 1
                self._edit = self._edit[:-1]
            return

        if self.stop is not None:
            # Use wants to try again immediately after score
            self.reset()
            head = self.get_stats(self.elapsed)
            self.screen.rerender_race(head)
            self.screen.show_keystroke(head,
                                       self.position,
                                       self.incorrect,
                                       self._edit,
                                       key)

        # Start recording upon first ordinary key press
        if self.start is None:
            self.start = time.time()

        if key == curses.KEY_ENTER:
            key = "\n"
        elif key == "\t" and self.tab_spaces is not None:
            key = " "*self.tab_spaces

        # Did the user strike the correct key?
        if self.incorrect == 0 and self.quote.text[self.position] == key:
            self.position += 1

            # Reset edit buffer on a correctly finished word
            if key == " " or key == "\n":
                self.screen.clear_prompt()
                self._edit = ""
            else:
                self._edit += key

            # Finished typing?
            if self.position == len(self.quote.text):
                self.mark_finished()
        elif self.incorrect + self.position < len(self.quote.text):
            self.incorrect += 1
            self.total_incorrect += 1
            if key == "\n":
                key = " "
            self._edit += key



"""
Difficulty scores for the built-in database.

Distributed under the GNU Affero General Public License (AGPL) v3 or later. See
the file LICENSE.txt for the full license text. This software makes use of open
source software.

The quotes database is *not* covered by the AGPL!
"""

import gzip
import pickle

import pkg_resources

class Difficulty(object):
    """Loads difficulty scores."""

    @staticmethod
    def _filename():
        """Returns the filename of the packaged difficulty database."""
        return pkg_resources.resource_filename("wpm",
                                               "data/difficulty.pickle.gz")

    @staticmethod
    def _normalize(diffs):
        """Normalizes difficulty scores to 0.0-1.0 where 1.0 is hardest."""
        low = min(diffs.values())
        delta = max(diffs.values()) - low
        out = {}

        for text_id, score in diffs.items():
            out[text_id] = 1.0 - (float(score - low) / delta)

        return out

    @staticmethod
    def load():
        """Loads the dictionary that maps text_id to difficulty scores.

        The difficulty scores are normalized and goes from 0.0 (easy) to 1.0
        (hard).
        """
        filename = Difficulty._filename()
        with gzip.open(filename, "rb") as file_obj:
            return Difficulty._normalize(pickle.load(file_obj))



# -*- encoding: utf-8 -*-

"""
This file is part of the wpm software.
Copyright 2017, 2018 Christian Stigen Larsen

Distributed under the GNU Affero General Public License (AGPL) v3 or later. See
the file LICENSE.txt for the full license text. This software makes use of open
source software.

The quotes database is *not* covered by the AGPL!
"""

import collections

class Recorder(object):
    """Class for recording keystrokes."""
    def __init__(self):
        self.reset()

    def add(self, elapsed, key, position, incorrect):
        """Adds a time stamp."""
        self.elapsed.append(elapsed)
        self.keys.append(key)
        self.position.append((position, incorrect))

    def reset(self):
        """Destroys all time stamps."""
        self.elapsed = collections.deque()
        self.keys = collections.deque()
        self.position = collections.deque()

    def __getitem__(self, index):
        elapsed = self.elapsed[index]
        key = self.keys[index]
        position, incorrect = self.position[index]
        return elapsed, key, position, incorrect

    def __len__(self):
        return len(self.elapsed)

class Playback(object):
    def __init__(self, recorder):
        self.recorder = recorder
        self.index = 0

    def next(self):
        values = self.recorder[self.index]
        self.index = (self.index + 1) % len(self.recorder)
        return values



# -*- encoding: utf-8 -*-

"""
This file is part of the wpm software.
Copyright 2017, 2018 Christian Stigen Larsen

Distributed under the GNU Affero General Public License (AGPL) v3 or later. See
the file LICENSE.txt for the full license text. This software makes use of open
source software.

The quotes database is *not* covered by the AGPL!
"""

import curses
import curses.ascii
import locale
import os
import sys

from wpm.config import Config
from wpm.error import WpmError
from wpm.gauss import confidence_interval

class Screen(object):
    """Renders the terminal screen."""

    COLOR_AUTHOR = 1
    COLOR_BACKGROUND = 2
    COLOR_CORRECT = 3
    COLOR_HISCORE = 4
    COLOR_INCORRECT = 5
    COLOR_PROMPT = 6
    COLOR_QUOTE = 7
    COLOR_STATUS = 8

    def __init__(self):
        self.config = Config()

        # Make delay slower
        os.environ.setdefault("ESCDELAY", self.config.escdelay)

        # Use the preferred system encoding
        locale.setlocale(locale.LC_ALL, "")
        self.encoding = locale.getpreferredencoding().lower()

        self.screen = curses.initscr()
        self.screen.nodelay(True)

        min_lines = 12
        if self.lines < min_lines:
            curses.endwin()
            raise WpmError(
                "wpm requires at least %d lines in your display" % min_lines)

        min_cols = 20
        if self.columns < min_cols:
            curses.endwin()
            raise WpmError(
                "wpm requires at least %d columns in your display" % min_cols)

        try:
            self.screen.keypad(True)
            curses.noecho()
            curses.cbreak()

            curses.start_color()
            self.set_colors()

            self.window = curses.newwin(self.lines, self.columns, 0, 0)
            self.window.keypad(True)
            self.window.timeout(self.config.window_timeout)
            self.window.bkgd(" ", Screen.COLOR_BACKGROUND)


            # Local variables related to quote. TODO: Move this mess to somewhere
            # else.
            self.cheight = 0
            self.first_key = True
            self.quote = ""
            self.quote_author = ""
            self.quote_columns = 0
            self.quote_coords = tuple()
            self.quote_height = 0
            self.quote_id = 0
            self.quote_lengths = tuple()
            self.quote_title = ""
        except:
            curses.endwin()
            raise

    @staticmethod
    def _word_wrap(text, width):
        """Returns lengths of lines that can be printed without wrapping."""
        lengths = []
        while len(text) > width:
            try:
                end = text[:width + 1].rindex(" ")
            except ValueError:
                break

            # We can't divide the input nicely, so just display it as-is
            if end == -1:
                return [len(text)]

            lengths.append(end)
            text = text[end + 1:]

        if text:
            lengths.append(len(text))

        return lengths

    @staticmethod
    def _screen_coords(lengths, position):
        """Translates quote offset into screen coordinates.

        Args:
            lengths: List of line lengths for the word-wrapped quote.
            position: Offset into the quote that we want to translate to screen
                      coordinates.

        Returns:
            Tuple containing X and Y screen coordinates.
        """
        y_position = 0

        for y_position, line_length in enumerate(lengths):
            if position <= line_length:
                break
            position -= line_length + 1

        return position, y_position



    @property
    def columns(self):
        """Returns number of terminal columns."""
        # pylint: disable=no-member
        return curses.COLS

    @property
    def lines(self):
        """Returns number of terminal lines."""
        # pylint: disable=no-member
        return curses.LINES

    def set_colors(self):
        """Sets up curses color pairs."""
        hicolor = os.getenv("TERM").endswith("256color")

        if hicolor:
            bg_col = self.config.background_color_256
            curses.init_pair(Screen.COLOR_AUTHOR, *self.config.author_color_256)
            curses.init_pair(Screen.COLOR_BACKGROUND, bg_col, bg_col)
            curses.init_pair(Screen.COLOR_CORRECT, *self.config.correct_color_256)
            curses.init_pair(Screen.COLOR_HISCORE, *self.config.score_highlight_color_256)
            curses.init_pair(Screen.COLOR_INCORRECT, *self.config.incorrect_color_256)
            curses.init_pair(Screen.COLOR_PROMPT, *self.config.prompt_color_256)
            curses.init_pair(Screen.COLOR_QUOTE, *self.config.quote_color_256)
            curses.init_pair(Screen.COLOR_STATUS, *self.config.status_color_256)
        else:
            bg_col = self.config.background_color
            curses.init_pair(Screen.COLOR_AUTHOR, *self.config.author_color)
            curses.init_pair(Screen.COLOR_BACKGROUND, bg_col, bg_col)
            curses.init_pair(Screen.COLOR_CORRECT, *self.config.correct_color)
            curses.init_pair(Screen.COLOR_HISCORE, *self.config.score_highlight_color)
            curses.init_pair(Screen.COLOR_INCORRECT, *self.config.incorrect_color)
            curses.init_pair(Screen.COLOR_PROMPT, *self.config.prompt_color)
            curses.init_pair(Screen.COLOR_QUOTE, *self.config.quote_color)
            curses.init_pair(Screen.COLOR_STATUS, *self.config.status_color)

        # Rebind class variables
        Screen.COLOR_AUTHOR = curses.color_pair(Screen.COLOR_AUTHOR)
        Screen.COLOR_BACKGROUND = curses.color_pair(Screen.COLOR_BACKGROUND)
        Screen.COLOR_CORRECT = curses.color_pair(Screen.COLOR_CORRECT)
        Screen.COLOR_HISCORE = curses.color_pair(Screen.COLOR_HISCORE)
        Screen.COLOR_INCORRECT = curses.color_pair(Screen.COLOR_INCORRECT)
        Screen.COLOR_PROMPT = curses.color_pair(Screen.COLOR_PROMPT)
        Screen.COLOR_QUOTE = curses.color_pair(Screen.COLOR_QUOTE)
        Screen.COLOR_STATUS = curses.color_pair(Screen.COLOR_STATUS)

        if not hicolor:
            # Make certain colors more visible
            Screen.COLOR_CORRECT |= curses.A_DIM
            Screen.COLOR_INCORRECT |= curses.A_UNDERLINE | curses.A_BOLD
            Screen.COLOR_QUOTE |= curses.A_BOLD
            Screen.COLOR_STATUS |= curses.A_BOLD

    @staticmethod
    def is_escape(key):
        """Checks for escape key."""
        if len(key) == 1:
            return ord(key) == curses.ascii.ESC
        return False

    @staticmethod
    def is_backspace(key):
        """Checks for backspace key."""
        if len(key) > 1:
            return key == "KEY_BACKSPACE"
        elif ord(key) in (curses.ascii.BS, curses.ascii.DEL):
            return True
        return False

    def get_key(self):
        """Gets a single key stroke."""
        # pylint: disable=method-hidden
        # Install a suitable get_key based on Python version
        if sys.version_info[0:2] >= (3, 3):
            self.get_key = self._get_key_py33
        else:
            self.get_key = self._get_key_py27
        return self.get_key()

    def _get_key_py33(self):
        """Python 3.3+ implementation of get_key."""
        # pylint: disable=too-many-return-statements
        try:
            # Curses in Python 3.3 handles unicode via get_wch
            key = self.window.get_wch()
            if isinstance(key, int):
                if key == curses.KEY_BACKSPACE:
                    return "KEY_BACKSPACE"
                elif key == curses.KEY_LEFT:
                    return "KEY_LEFT"
                elif key == curses.KEY_RIGHT:
                    return "KEY_RIGHT"
                elif key == curses.KEY_RESIZE:
                    return "KEY_RESIZE"
                return None
            return key
        except curses.error:
            return None
        except KeyboardInterrupt:
            raise

    def _get_key_py27(self):
        """Python 2.7 implementation of get_key."""
        # pylint: disable=too-many-return-statements
        try:
            key = self.window.getkey()

            # Start of UTF-8 multi-byte character?
            if self.encoding == "utf-8" and ord(key[0]) & 0x80:
                multibyte = key[0]
                cont_bytes = ord(key[0]) << 1
                while cont_bytes & 0x80:
                    cont_bytes <<= 1
                    multibyte += self.window.getkey()[0]
                return multibyte.decode(self.encoding)

            if isinstance(key, int):
                if key == curses.KEY_BACKSPACE:
                    return "KEY_BACKSPACE"
                elif key == curses.KEY_LEFT:
                    return "KEY_LEFT"
                elif key == curses.KEY_RIGHT:
                    return "KEY_RIGHT"
                elif key == curses.KEY_RESIZE:
                    return "KEY_RESIZE"
                return None
            return key.decode("ascii")
        except KeyboardInterrupt:
            raise
        except curses.error:
            return None

    def addstr(self, x_pos, y_pos, text, color=None):
        """Wraps call around curses.window.addsr."""
        if self.lines > y_pos >= 0:
            if x_pos >= 0 and (x_pos + len(text)) < self.columns:
                self.window.addstr(y_pos, x_pos, text, color)

    def chgat(self, x_pos, y_pos, length, color):
        """Wraps call around curses.window.chgat."""
        if self.lines > y_pos >= 0:
            if x_pos >= 0 and (x_pos + length) <= self.columns:
                self.window.chgat(y_pos, x_pos, length, color)

    def set_cursor(self, x_pos, y_pos):
        """Sets cursor position."""
        if (y_pos < self.lines) and (x_pos < self.columns):
            self.window.move(y_pos, x_pos)

    def right_column(self, y_pos, x_pos, width, text):
        """Writes text to screen in coumns."""
        lengths = Screen._word_wrap(text, width)

        for cur_y, length in enumerate(lengths, y_pos):
            self.addstr(x_pos - length, cur_y,
                        text[:length].encode(self.encoding),
                        Screen.COLOR_AUTHOR)
            text = text[1+length:]

        return len(lengths)

    def update_quote(self, color):
        """Renders complete quote on screen."""
        quote = self.quote[:]
        for y_pos, length in enumerate(self.quote_lengths, 2):
            self.addstr(0, y_pos, quote[:length].encode(self.encoding), color)
            quote = quote[1 + length:]

    def update_author(self):
        """Renders author on screen."""
        author = u"— %s, %s" % (self.quote_author, self.quote_title)
        self.cheight = 4 + self.quote_height
        self.cheight += self.right_column(self.cheight - 1,
                                          self.quote_columns - 10,
                                          self.quote_columns // 2,
                                          author)

    def update_header(self, text):
        """Renders top-bar header."""
        self.addstr(0, 0, text, Screen.COLOR_STATUS)
        self.chgat(0, 0, self.columns, Screen.COLOR_STATUS)
        #self.window.chgat(0, 0, self.columns, Screen.COLOR_STATUS)

    def set_quote(self, quote):
        """Sets up variables used for a new quote."""
        # TODO: Move this stuff elsewhere
        if self.config.max_quote_width > 0:
            self.quote_columns = min(self.columns, self.config.max_quote_width)
        else:
            self.quote_columns = self.columns

        self.cheight = 0
        self.quote = quote.text
        self.quote_author = quote.author
        self.quote_id = quote.text_id
        self.quote_title = quote.title
        self.quote_lengths = tuple(Screen._word_wrap(self.quote,
                                                     self.quote_columns - 1))
        self.quote_height = len(self.quote_lengths)

        # Remember (x, y) position for each quote offset.
        self.quote_coords = []
        for offset in range(len(self.quote)+1):
            x_pos, y_pos = Screen._screen_coords(self.quote_lengths, offset)
            self.quote_coords.append((x_pos, y_pos))
        self.quote_coords = tuple(self.quote_coords)

    def clear_prompt(self):
        self.set_cursor(0, self.cheight)
        self.window.clrtoeol()

    def update_prompt(self, prompt):
        """Prints prompt on the display."""
        self.addstr(0, self.cheight, (prompt + " ").encode(self.encoding),
                    Screen.COLOR_PROMPT)

    def show_browser(self, head, stats):
        """Show quote browsing screen."""
        self.update_header(head)
        self.update_quote(Screen.COLOR_QUOTE)
        self.update_author()
        self.show_help()
        self.show_stats(stats)
        self.set_cursor(0, 2)

    def show_help(self):
        """Shows help instructions on screen."""
        self.cheight += 1
        self.set_cursor(0, self.cheight)
        self.addstr(0, self.cheight,
                    "Start typing, hit SPACE/ARROWS to browse or ESC to quit.",
                    Screen.COLOR_PROMPT)

    def show_stats(self, stats):
        """Shows statistics for the current quote."""
        results = stats.text_id_results(stats.keyboard, self.quote_id)

        if len(results) < 2:
            return

        percent = self.config.confidence_interval_percent
        alpha = 1 - (percent/100.0)
        samples = len(results)

        wpm_avg, acc_avg = results.averages()
        wpm_sd, acc_sd = results.stddevs()
        wpm_min, wpm_max, acc_min, acc_max = results.extremals()
        wpm_ci0, wpm_ci1 = confidence_interval(wpm_avg, wpm_sd, samples, alpha)
        acc_ci0, acc_ci1 = confidence_interval(acc_avg, acc_sd, samples, alpha)

        msg = "wpm %5.1f min %5.1f avg %5.1f max %5.1f sd [%5.1f-%5.1f] %d%% ci (n=%d)" % (
                wpm_min, wpm_avg, wpm_max, wpm_sd, wpm_ci0, wpm_ci1, percent,
                samples)
        self.cheight += 2
        self.addstr(0, self.cheight, msg, Screen.COLOR_CORRECT)

        msg = "acc %5.1f min %5.1f avg %5.1f max %5.1f sd [%5.1f %5.1f] %d%% ci (n=%d)" % (
                100*acc_min, 100*acc_avg, 100*acc_max, 100*acc_sd, 100*acc_ci0,
                100*acc_ci1, percent, samples)
        self.cheight += 1
        self.addstr(0, self.cheight, msg, Screen.COLOR_CORRECT)
        self.cheight += 1

    def show_score(self, head, wpm_score, stats):
        """Show score screen after typing has finished."""
        self.update_header(head)
        self.update_quote(Screen.COLOR_CORRECT)
        self.update_author()

        # Highlight score
        score = "You scored %.1f wpm!" % wpm_score
        self.update_prompt(score)
        if len(score) < self.columns:
            self.chgat(11, self.cheight, len(str("%.1f" % wpm_score)),
                       Screen.COLOR_HISCORE)

        self.show_help()
        self.show_stats(stats)
        self.set_cursor(0, 2)

    def highlight_progress(self, position, incorrect):
        """Colors finished and incorrectly typed parts of the quote."""
        if incorrect:
            color = Screen.COLOR_INCORRECT
        else:
            color = Screen.COLOR_CORRECT

        # Highlight correct / incorrect character in quote
        ixpos, iypos = self.quote_coords[position + incorrect - 1]
        color = Screen.COLOR_INCORRECT if incorrect else Screen.COLOR_CORRECT
        self.chgat(max(ixpos, 0), 2 + iypos, 1, color)

        # Highlight next as correct, in case of backspace
        xpos, ypos = self.quote_coords[position + incorrect]
        self.chgat(xpos, 2 + ypos, 1, Screen.COLOR_QUOTE)

    def show_keystroke(self, head, position, incorrect, typed, key):
        """Updates the screen while typing."""
        self.update_header(head)

        if key and (position + incorrect) <= len(self.quote):
            self.highlight_progress(position, incorrect)
            self.update_prompt("> " + typed)

        # Move cursor to current position in text before refreshing
        xpos, ypos = self.quote_coords[position + incorrect]
        self.set_cursor(min(xpos, self.quote_columns - 1), 2 + ypos)

    def rerender_race(self, head):
        """Re-renders currently running game."""
        self.clear()
        self.update_header(head)
        self.update_quote(Screen.COLOR_QUOTE)
        self.update_author()
        self.set_cursor(0, 2)

    def clear(self):
        """Clears the screen."""
        self.window.clear()

    def deinit(self):
        """Deinitializes curses."""
        curses.nocbreak()
        self.screen.keypad(False)
        curses.echo()
        curses.endwin()



"""
Setup script for WPM.

This file is part of the wpm software.
Copyright 2017, 2018 Christian Stigen Larsen

Distributed under the GNU Affero General Public License (AGPL) v3 or later. See
the file LICENSE.txt for the full license text. This software makes use of open
source software.

The quotes database is *not* covered by the AGPL!
"""

import os
import runpy

from setuptools import setup

def get_version():
    """Reads current WPM version from disk."""
    filename = os.path.join(os.path.dirname(__file__), "wpm", "__init__.py")
    var = runpy.run_path(filename)
    return var["__version__"]

_VERSION = get_version()

setup(
    name="wpm",
    scripts=["scripts/wpm"],
    version=_VERSION,
    description="Console app for measuring typing speed in words per minute (WPM)",
    author="Christian Stigen Larsen",
    author_email="csl@csl.name",
    packages=["wpm"],
    package_dir={"wpm": "wpm"},
    package_data={"wpm": ["data/examples.json.gz"]},
    include_package_data=True,
    url="https://github.com/cslarsen/wpm",
    download_url="https://github.com/cslarsen/wpm/tarball/v%s" % _VERSION,
    license="https://www.gnu.org/licenses/agpl-3.0.html",
    long_description=open("README.rst").read(),
    zip_safe=True,
    test_suite="tests",
    keywords=["wpm", "typing", "typist"],
    platforms=["unix", "linux", "osx", "cygwin"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.1",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
)



import os
import unittest

from wpm.error import WpmError
from wpm.quotes import Quotes, Quote


class QuotesTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.quotes = Quotes.load()

    def test_default_database(self):
        self.assertGreaterEqual(len(self.quotes), 4000)

    def test_from_id(self):
        with self.assertRaises(KeyError):
            self.quotes.from_id(-1)

        text_id = 3621031
        quote = self.quotes.from_id(text_id)
        self.assertEqual(quote.text_id, text_id)
        self.assertEqual(quote.author, "Joseph Heller")
        self.assertEqual(quote.title, "Catch-22")
        self.assertTrue(quote.text.startswith("Let's take a drive"))

    def test_load_json(self):
        filename = os.path.join(os.path.dirname(__file__), "test.json")

        quotes = Quotes.load_json(filename)
        self.assertEqual(len(quotes), 1)

        quote = quotes.at(0)
        self.assertEqual(quote.author, "Mr. Author")
        self.assertEqual(quote.title, "The Title")
        self.assertEqual(quote.text, "This is the text.")

    def test_load_json_missing_file(self):
        with self.assertRaises(WpmError):
            Quotes.load_json("non-existing-file")

    def test_load_json_invalid_format(self):
        with self.assertRaises(WpmError):
            Quotes.load_json(__file__)


class QuoteTests(unittest.TestCase):
    def test_from_tuple(self):
        author = "The Author"
        title = "The Title"
        text = "The Text."
        text_id = 123

        quote = Quote.from_tuple((author, title, text, text_id))

        self.assertEqual(quote.author, author)
        self.assertEqual(quote.title, title)
        self.assertEqual(quote.text, text)
        self.assertEqual(quote.text_id, text_id)
